#include "config.h"

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>

#include <syslog.h>

#include "gdm.h"
#include "vicious.h"

#include "greeter_parser.h"
#include "greeter_configuration.h"
#include "greeter_item_throbber.h"
#include "greeter_events.h"


typedef struct
{
  int n_steps;
  int step;
  GdkPixbuf **pixbufs;
  GreeterItemInfo *info;
} Throbber;

static guint throbber_handler_id = 0;
static Throbber *throbber = NULL;
static gchar *datadir = NULL;

static gboolean
throbber_timer (gpointer data)
{
  Throbber *throbber = data;

  if (++throbber->step == throbber->n_steps)
    throbber->step = 0;
  
  g_object_unref (throbber->info->data.pixmap.pixbufs[GREETER_ITEM_STATE_NORMAL]);
  throbber->info->data.pixmap.pixbufs[GREETER_ITEM_STATE_NORMAL] = g_object_ref (throbber->pixbufs[throbber->step]);

  gnome_canvas_item_set (throbber->info->item,
                         "pixbuf", throbber->pixbufs[throbber->step],                                    NULL);
  
  return TRUE;
}

static void
setup_throbber (Throbber *throbber, gboolean warning) 
{
  GdkPixbuf *pixbuf;
  gint i, size;
  gchar *file;

  throbber->n_steps = 0;

  file = g_build_filename (datadir ? datadir : "/usr/share/gdm/themes/Bluecurve",
                           warning ? "throbber-warning.png" : "throbber.png", 
                           NULL);

  pixbuf = gdk_pixbuf_new_from_file (file, NULL);

  g_free (file);

  if (pixbuf == NULL) 
 	return;

  for (i = 0; i < throbber->n_steps; i++)
    {
       g_object_unref (throbber->pixbufs[i]); 
       throbber->pixbufs[i] = NULL;
    }
  g_free (throbber->pixbufs);

  size = gdk_pixbuf_get_height (pixbuf);
  throbber->n_steps = gdk_pixbuf_get_width (pixbuf)/size;
  throbber->pixbufs = g_malloc (sizeof (GdkPixbuf *) * throbber->n_steps);

  for (i = 0; i < throbber->n_steps; i++)
    throbber->pixbufs[i] = gdk_pixbuf_new_subpixbuf (pixbuf,
					       	     i*size, 0,
						     size, size);

  throbber->step = throbber->step % throbber->n_steps;
}

static void
set_throbber_visible (gboolean visible)
{
  static const gchar *id[] = {
    "throbber-background",
    "throbber-label",
    "throbber-pixmap",
    NULL
  };

  GreeterItemInfo *info;
  gint i;

  for (i = 0; id[i]; i++) {
    info = greeter_lookup_id (id[i]);
    if (info) {
      if (visible) {
        if (info->item)
          gnome_canvas_item_show (GNOME_CANVAS_ITEM (info->item));
        if (info->group_item)
          gnome_canvas_item_show (GNOME_CANVAS_ITEM (info->group_item));
      }
      else {
        if (info->item)
          gnome_canvas_item_hide (GNOME_CANVAS_ITEM (info->item));
        if (info->group_item)
          gnome_canvas_item_hide (GNOME_CANVAS_ITEM (info->group_item));
      }
    } 
  }
}

static void 
greeter_throbber_handler (GreeterItemInfo *info,
                          gpointer         user_data)
{ 
#if 0
  gdk_spawn_command_line_on_screen (gdk_screen_get_default (),
                                    "/usr/bin/rhgb-client --details=yes", 
                                    NULL);
#endif
}

gboolean 
greeter_item_throbber_setup (const char *theme_dir)
{

  g_free (datadir);
  datadir = g_strdup (theme_dir);

  set_throbber_visible (FALSE);
}

void
greeter_item_throbber_start (gboolean warning)
{
  GreeterItemInfo *info;

  info = greeter_lookup_id ("throbber-pixmap");
  if (info) 
    {
      if (throbber == NULL) 
        {
           throbber = g_new0 (Throbber, 1);
           throbber->info = info;
           throbber->n_steps = 0;
           throbber->step = 0;

           greeter_item_register_action_callback ("throbber",
                                                  (ActionFunc)greeter_throbber_handler,
                                                  NULL);
        }

      setup_throbber (throbber, warning);

      if (throbber->n_steps == 0)
	return;

      if (throbber_handler_id == 0)
        throbber_handler_id = g_timeout_add (83, throbber_timer, throbber);
    }

  set_throbber_visible (TRUE);

#if 0
  gdk_spawn_command_line_on_screen (gdk_screen_get_default (),
                                    "/usr/bin/rhgb -f -u :0 --show-details", 
                                    NULL);

#endif
}

void
greeter_item_throbber_stop (void)
{
  GreeterItemInfo *info;
  gint i;

  set_throbber_visible (FALSE);

  if (throbber_handler_id != 0)
    {
      g_source_remove (throbber_handler_id);
      throbber_handler_id = 0;

      for (i = 0; i < throbber->n_steps; i++)
        g_object_unref (throbber->pixbufs[i]); 
      
      g_free (throbber);
      throbber = NULL;
    }
}

